##########################################################################################
# tests/test_vicarlabel.py
##########################################################################################

import os
import pathlib
import shutil
import sys
import unittest
from vicar.vicarlabel import VicarLabel, VicarError, _REQUIRED
from vicar._LABEL_GRAMMAR import _LABEL_GRAMMAR


class Test_VicarLabel(unittest.TestCase):

    def test_VicarLabelGO(self):

        vicar_dir = pathlib.Path(sys.modules['vicar'].__file__)
        test_dir = vicar_dir.parent.parent / 'test_files'

        text = (
            "LBLSIZE=1536            FORMAT='BYTE'  TYPE='TABULAR'  BUFSIZ=20480  "
            "DIM=3  EOL=1  RECSIZE=512  ORG='BSQ'    NS=512  NB=1  N1=512  N2=1  "
            "N3=1  N4=0  NBB=0    HOST='AXP-VMS'  INTFMT='LOW'  "
            "REALFMT='VAX'          NL=0            NLB=18  BHOST='AXP-VMS'  "
            "BINTFMT='LOW'  BREALFMT='VAX'  BLTYPE='IBIS'  "
            "PROPERTY='IBIS'                TYPE='TIEPOINT'  NR=552  NC=4  ORG='ROW'  "
            "FMT_DEFAULT='REAL'  GROUPS=('LINE','SAMP','C_POS_IMAGE','INPUT','POSITION',"
            "'C_POSITION','PIXEL','C_PIXEL','OUTPUT','C_POINT','C_ROOT')  GROUP_1=(3,1)  "
            "GROUP_2=(4,2)  GROUP_3=(3,4,1,2)  GROUP_4=(3,4)  GROUP_5=(1,2,3,4)  "
            "GROUP_6=(3,4,1,2)  GROUP_7=(1,2,3,4)  GROUP_8=(1,2,3,4)  GROUP_9=(1,2)  "
            "GROUP_10=(1,2,3,4)  GROUP_11=(3,4,1,2)  SEGMENT=16  BLOCKSIZE=512  "
            "COFFSET=(0,4,8,12)  PROPERTY='TIEPOINT'  NUMBER_OF_AREAS_HORIZONTAL=23  "
            "NUMBER_OF_AREAS_VERTICAL=22  TASK='TASK'  USER='SHOWALTER'  "
            "DAT_TIM='Sun Oct  2 05:05:17 2011'  "
            "LAB01='                     800     800 800 800 L 1                          SC'  "
            "LAB02='VGR-2   FDS 20693.02   PICNO 0215J2+001   SCET 79.192 01:19:58         C'  "
            "LAB03='WA CAMERA  EXP   15360.0 MSEC FILT 2(CLEAR )  LO GAIN  SCAN RATE  5:1  C'  "
            "LAB04='ERT 79.192 02:11:56   1/ 2 FULL    RES   VIDICON TEMP  -80.00 DEG C    C'  "
            "LAB05='IN/205140/14 OUT/xxxxxx/xx     J_RINGS     DSS #14   BIT SNR    6.273  C'  "
            "LAB06=' xxxxx A/xxxxxxxx B/xxxx C/xxxx D/xxxxxxxx ETLM/xxxxxxxxxxxxxxxxxxxxS AC'  "
            "LBLSIZE=1024            "
            "LAB07='NA OPCAL xx(015360.0*MSEC)PIXAVG 032/0 OPERATIONAL MODE 3(WAONLY)     AC'  "
            "LAB08='CAM ECAL CYCLE BEAM  RESET OPEN  CLOSE FLOOD AEXPM  FIL G1 SHUT MODE  AC'  "
            "LAB09='NA   NO   PREP  NO    YES   NO    NO    NO    NO    0 P  * NORMAL     AC'  "
            "LAB10='WA   NO   READ  YES   NO    NO    NO    NO    NO    2 P  7 NORMAL     AC'  "
            "LAB11='LSB_TRUNC=OFF  TLM_MODE=IM-2D COMPRESSION=OFF                          L'  "
            "NLABS=11    TASK='VGRFILLI'  USER='SHOWALTER'  DAT_TIM='Sun Oct  2 05:05:17 2011'  "
            "LIN_CNT=0  TASK='RESLOC'  USER='SHOWALTER'  DAT_TIM='Sun Oct  2 05:05:18 2011'    "
        )

        def test_vic(source):

            vic = VicarLabel(source)
            self.assertEqual(vic['LBLSIZE'], 1536)
            self.assertEqual(vic[('LBLSIZE',0)], 1536)
            self.assertEqual(vic[('LBLSIZE',1)], 1024)
            self.assertEqual(vic[('LBLSIZE',-1)], 1024)
            self.assertEqual(vic[('LBLSIZE',-2)], 1536)
            self.assertEqual(vic['LBLSIZE+'], [1536,1024])
            self.assertRaises(KeyError, vic.__getitem__, 'FOO')
            self.assertRaises(KeyError, vic.__getitem__, 'FOO+')
            self.assertRaises(TypeError, vic.__getitem__, 3.14159)
            self.assertRaises(IndexError, vic.__getitem__, ('LBLSIZE',2))
            self.assertRaises(TypeError, vic.__getitem__, set())

            self.assertIn('LBLSIZE', vic)
            self.assertIn(('LBLSIZE',0), vic)
            self.assertIn(('LBLSIZE',1), vic)
            self.assertNotIn(('LBLSIZE',2), vic)
            self.assertIn(('LBLSIZE',-1), vic)
            self.assertNotIn(('LBLSIZE',-3), vic)
            self.assertNotIn('FOO', vic)

            self.assertEqual(vic[0], 1536)
            self.assertEqual(vic[-2], 'SHOWALTER')
            self.assertRaises(IndexError, vic.__getitem__, 999)

            self.assertEqual(vic.get('LBLSIZE', -1), 1536)
            self.assertEqual(vic.get(('LBLSIZE',0), -1), 1536)
            self.assertEqual(vic.get(('LBLSIZE',1), -1), 1024)
            self.assertEqual(vic.get(('LBLSIZE',2), -1), -1)

            self.assertEqual(len(vic), 71)

            self.assertEqual(vic.arg('LBLSIZE'), 0)
            self.assertEqual(vic.arg('DIM'), 4)
            self.assertEqual(vic.arg(('LBLSIZE',1)), 57)

            self.assertEqual(vic.arg(0), 0)
            self.assertEqual(vic.arg(4), 4)
            self.assertEqual(vic.arg(-1), 70)
            self.assertEqual(vic.arg(-71), 0)
            self.assertRaises(IndexError, vic.arg, 71)
            self.assertRaises(IndexError, vic.arg, -72)
            self.assertRaises(IndexError, vic.arg, ('LBLSIZE',2))
            self.assertRaises(KeyError, vic.arg, ('FOO',0))

            # Iterators

            self.assertEqual(list(vic)[:3], [('LBLSIZE',0), 'FORMAT', ('TYPE',0)])

            self.assertEqual(list(vic.names('TASK')), ['TASK', 'TASK', 'TASK'])
            self.assertEqual(list(vic.names())[:3], ['LBLSIZE', 'FORMAT', 'TYPE'])

            self.assertEqual(list(vic.keys('TASK')), [('TASK',0), ('TASK',1), ('TASK',2)])
            self.assertEqual(list(vic.keys())[:3], [('LBLSIZE',0), 'FORMAT', ('TYPE',0)])

            self.assertEqual(list(vic.values('TASK')), ['TASK', 'VGRFILLI', 'RESLOC'])
            self.assertEqual(list(vic.values())[:3], [1536, 'BYTE', 'TABULAR'])

            self.assertEqual(list(vic.items('TASK', unique=False)),
                             [('TASK','TASK'), ('TASK','VGRFILLI'), ('TASK','RESLOC')])
            self.assertEqual(list(vic.items('TASK', unique=True)),
                             [(('TASK',0), 'TASK'), (('TASK',1), 'VGRFILLI'),
                              (('TASK',2), 'RESLOC')])
            self.assertEqual(list(vic.items(unique=False))[:3],
                             [('LBLSIZE', 1536), ('FORMAT', 'BYTE'), ('TYPE', 'TABULAR')])
            self.assertEqual(list(vic.items(unique=True))[:3],
                             [(('LBLSIZE',0), 1536), ('FORMAT', 'BYTE'),
                              (('TYPE',0), 'TABULAR')])

            self.assertEqual(list(vic.items(r'GROUP_\d+')), [('GROUP_1', [3, 1]),
                                                             ('GROUP_2', [4, 2]),
                                                             ('GROUP_3', [3, 4, 1, 2]),
                                                             ('GROUP_4', [3, 4]),
                                                             ('GROUP_5', [1, 2, 3, 4]),
                                                             ('GROUP_6', [3, 4, 1, 2]),
                                                             ('GROUP_7', [1, 2, 3, 4]),
                                                             ('GROUP_8', [1, 2, 3, 4]),
                                                             ('GROUP_9', [1, 2]),
                                                             ('GROUP_10', [1, 2, 3, 4]),
                                                             ('GROUP_11', [3, 4, 1, 2])])
            self.assertEqual(list(vic.values(r'GROUP_\d+')), [[3, 1],
                                                              [4, 2],
                                                              [3, 4, 1, 2],
                                                              [3, 4],
                                                              [1, 2, 3, 4],
                                                              [3, 4, 1, 2],
                                                              [1, 2, 3, 4],
                                                              [1, 2, 3, 4],
                                                              [1, 2],
                                                              [1, 2, 3, 4],
                                                              [3, 4, 1, 2]])

            self.assertEqual(list(vic.args(r'GROUP_\d+')), list(range(31,42)))
            self.assertEqual(list(vic.args()), list(range(71)))

        # Alternative sources: text string, file, list of tuples
        test_vic(text)
        test_vic(test_dir / 'C2069302_GEOMA.DAT')
        test_vic(str(test_dir / 'C2069302_GEOMA.DAT'))

        with open(test_dir / 'C2069302_GEOMA.DAT', 'rb') as f:
            test_vic(f)
            self.assertFalse(f.closed)
        self.assertTrue(f.closed)

        params = _LABEL_GRAMMAR.parse_string(text).as_list()
        test_vic(params)

        # __eq__
        reference = VicarLabel(text)
        self.assertEqual(reference, VicarLabel(test_dir / 'C2069302_GEOMA.DAT'))
        self.assertEqual(reference, VicarLabel(str(test_dir / 'C2069302_GEOMA.DAT')))
        self.assertEqual(reference, VicarLabel.from_file(test_dir / 'C2069302_GEOMA.DAT'))
        self.assertEqual(reference, VicarLabel(params))
        self.assertNotEqual(reference, set())

        # append
        parts = text.partition('LBLSIZE=1024')
        vic2 = VicarLabel(parts[0])
        vic2.append(parts[1] + parts[2])
        self.assertEqual(reference, vic2)

        vic3 = vic2.copy()
        vic2.append('PI=3.14159')
        vic3.append([('PI', 3.14159)])
        self.assertEqual(vic3, vic3)

        # skip over LBLSIZE and it will be filled in
        params = _LABEL_GRAMMAR.parse_string(parts[0]).as_list()
        missing_lblsize = VicarLabel(params[1:])
        self.assertEqual(missing_lblsize.names()[0], 'LBLSIZE')
        self.assertEqual(missing_lblsize['LBLSIZE'], 0)
        missing_lblsize[0] = (1536, 12)
        missing_lblsize.append(parts[1] + parts[2])
        self.assertEqual(reference, missing_lblsize)

        # export
        (label0, label1) = reference.export(resize=True)
        self.assertEqual(label1, '')
        self.assertRaises(IndexError, reference.__getitem__, ('LBLSIZE',1))
        self.assertEqual(len(label0), reference['LBLSIZE'])

        vic = VicarLabel(text)  # reinitialize
        (label0, label1) = vic.export(resize=False)
        self.assertEqual(len(label0), vic['LBLSIZE'])
        self.assertEqual(len(label1), vic[('LBLSIZE',1)])

        test = VicarLabel()
        test['RECSIZE'] = 100
        (label0, label1) = test.export(resize=False)
        self.assertEqual(len(label1), 0)    # resize switched to True

        test['RECSIZE'] = 200
        test['LBLSIZE'] = 200
        (label0, label1) = test.export(resize=False)
        self.assertEqual(len(label0), 200)
        self.assertGreater(len(label1), 0)  # resize preserved

        test['RECSIZE'] = 200
        test['LBLSIZE'] = 201
        (label0, label1) = test.export(resize=False)
        self.assertEqual(len(label1), 0)    # resize switched to True

        # as_string, __str__, __repr__
        test = VicarLabel()
        test.append("LBLSIZE=100  NOTE='more stuff'")
        # Fails on platforms other than Mac
#     self.assertEqual(test.as_string(),
#                      "LBLSIZE=0             FORMAT='BYTE'  TYPE='IMAGE'  BUFSIZ=20480  "
#                      "DIM=3  EOL=0  RECSIZE=0  ORG='BSQ'  NL=0  NS=0  NB=0  N1=0  N2=0  "
#                      "N3=0  N4=0  NBB=0  NLB=0  HOST='MAC-OSX'  INTFMT='LOW'  "
#                      "REALFMT='RIEEE'  BHOST='MAC-OSX'  BINTFMT='LOW'  BREALFMT='RIEEE'  "
#                      "BLTYPE=''  LBLSIZE=100           NOTE='more stuff'  ")
        self.assertEqual(test.as_string(stop=17),
                         "LBLSIZE=0             FORMAT='BYTE'  TYPE='IMAGE'  BUFSIZ=20480  "
                         "DIM=3  EOL=0  RECSIZE=0  ORG='BSQ'  NL=0  NS=0  NB=0  N1=0  N2=0  "
                         "N3=0  N4=0  NBB=0  NLB=0  ")
        self.assertEqual(test.as_string(start='BLTYPE'),
                         "BLTYPE=''  LBLSIZE=100           NOTE='more stuff'  ")
        self.assertEqual(test.as_string(start='BLTYPE', sep='xxx'),
                         "BLTYPE=''  xxxLBLSIZE=100           NOTE='more stuff'  ")
        self.assertEqual(test.as_string(start='BLTYPE', stop='NOTE', sep='xxx'),
                         "BLTYPE=''  xxxLBLSIZE=100           ")

        self.assertEqual(str(test), test.as_string())
        test2 = eval(repr(test))
        self.assertEqual(test, test2)

        # del of required
        test.append("ORG='ROWS'")
        with self.assertRaises(VicarError):
            del test['ORG']

        with self.assertRaises(VicarError):
            del test['NLB']

        with self.assertRaises(VicarError):
            del test['BLTYPE']

        del test['ORG',1]
        self.assertEqual(test2, test)

        # del
        vic = VicarLabel(text)  # reinitialize
        for k in range(1,11):
            del vic['LAB' + ('%02d' % k)]

        (label0, label1) = vic.export(resize=False)
        self.assertEqual(len(label1), 0)
        self.assertEqual(len(label0), vic['LBLSIZE'])
        self.assertEqual(len(vic['LBLSIZE+']), 1)

        # Restore LABxx
        vic2 = VicarLabel(text)
        for k in range(1,11):
            vic['LAB' + ('%02d' % k)] = vic2['LAB' + ('%02d' % k)]

        (label0, label1) = vic.export(resize=False)
        self.assertEqual(len(vic['LBLSIZE+']), 2)
        self.assertEqual(len(label1), vic['LBLSIZE+'][1])
        self.assertEqual(len(label0), vic['LBLSIZE'])

        # write_label
        dest = test_dir / 'C2069302_GEOMA_new_label.DAT'
        shutil.copy(test_dir / 'C2069302_GEOMA.DAT', dest)
        vic = VicarLabel(dest)
        for k in range(1,11):
            del vic['LAB' + ('%02d' % k)]
        vic.write_label(dest)

        altvic = VicarLabel(dest)
        self.assertEqual(vic, altvic)
        self.assertEqual(len(vic['LBLSIZE+']), 1)
        self.assertEqual(os.path.getsize(dest), vic['LBLSIZE'] + vic['RECSIZE'] * vic['NLB'])
        self.assertEqual(altvic.filepath, dest)

        vic2 = VicarLabel(text)
        for k in range(1,11):
            vic['LAB' + ('%02d' % k)] = vic2['LAB' + ('%02d' % k)]

        vic.write_label()

        altvic = VicarLabel(dest)
        self.assertEqual(vic, altvic)
        self.assertEqual(len(vic['LBLSIZE+']), 2)
        self.assertEqual(os.path.getsize(dest),
                         vic['LBLSIZE'] + vic['RECSIZE'] * vic['NLB'] + vic[('LBLSIZE',1)])

        altvic.filepath = None
        self.assertRaises(ValueError, altvic.write_label)

        altvic.filepath = str(dest)
        self.assertIsInstance(altvic.filepath, pathlib.Path)
        self.assertEqual(altvic.filepath, dest)

        os.remove(dest)     # delete extra file

        # Empty label
        empty = VicarLabel()
        self.assertEqual(len(empty), len(_REQUIRED))

        # Single tuple
        almost_empty = VicarLabel(('SEVEN', 7.0))
        self.assertEqual(len(almost_empty), len(_REQUIRED) + 1)
        almost_empty2 = VicarLabel({'SEVEN': (7.,)})
        self.assertEqual(almost_empty, almost_empty2)
        almost_empty2 = VicarLabel({'SEVEN': (7.,'')})
        self.assertEqual(almost_empty, almost_empty2)

        # Dict
        almost_empty2 = VicarLabel({'SEVEN': 7.})
        self.assertEqual(almost_empty, almost_empty2)
        almost_empty2 = VicarLabel({'SEVEN': (7.,)})
        self.assertEqual(almost_empty, almost_empty2)

        # Invalid inputs
        self.assertRaises(VicarError, VicarLabel,
                          "FORMAT='BYTE'  SEVEN=7.0  LBLSIZE=100  MORE=xyz")
        self.assertRaises(TypeError, VicarLabel, set([1,2,3]))
        self.assertRaises(TypeError, VicarLabel, [('LBLSIZE',100), set()])
        self.assertRaises(TypeError, VicarLabel, [('LBLSIZE',100), ('RECSIZE',)])
        self.assertRaises(TypeError, VicarLabel, [('LBLSIZE',100), 'RECSIZE'])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('TWO',2.,'',4,5,6,7)])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('TWO',2.,3,4,5,6)])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('TWO',2.,3,'')])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), (1,2.)])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('TWo',2.)])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('CR','\n')])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('LIST1',[])])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('LIST2',[1,2.])])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('SET',set())])
        self.assertRaises(VicarError, VicarLabel, [('LBLSIZE',100), ('SET',[set()])])
        self.assertRaises(VicarError, VicarLabel, [('A23456789012345678901234567890123',7)])
        test = VicarLabel([('A2345678901234567890123456789012',7)])
        test = VicarLabel([('LBLSIZE',100), ('TWo',2.), ('CR','\n'), ('LIST1',[]),
                           ('LIST2',[1,2.]), ('A23456789012345678901234567890123',7)],
                          strict=False)
        self.assertRaises(VicarError, VicarLabel,
                          [('LBLSIZE',100), ('SET',[1,2,set()])], strict=False)

        # Invalid formats
        self.assertRaises(TypeError, VicarLabel, [('ABC', 4, '%.4f')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 4, '%.4g')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 4, '%.4e')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 4., '%4d')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 4., '%4i')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 'abc', '%4i')])
        self.assertRaises(TypeError, VicarLabel, [('ABC', 'abc', '%s')])

        self.assertRaises(TypeError, VicarLabel, [('ABC', [(4, '%.4f')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [(4, '%.4g')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [(4, '%.4e')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [(4., '%4d')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [(4., '%4i')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [('abc', '%4i')])])
        self.assertRaises(TypeError, VicarLabel, [('ABC', [('abc', '%s')])])
        self.assertRaises(VicarError, VicarLabel, [('ABC', [(7, '%02d', 1, 2, 3)])])

        # Moving LBLSIZE to front
        vic = VicarLabel("FORMAT='BYTE'  SEVEN=7.0  LBLSIZE=100  MORE='LESS'")
        self.assertEqual(vic.items()[0], ('LBLSIZE', 100))

        # __setitem__, validation
        self.assertRaises(VicarError, vic.__setitem__, '_INVALID_NAME', 1)
        self.assertRaises(VicarError, vic.__setitem__, 'TUPLE', [])     # no empty lists
        self.assertRaises(VicarError, vic.__setitem__, 'SET', set())
        self.assertRaises(VicarError, vic.__setitem__, 'LIST', [1,2.])
        self.assertRaises(VicarError, vic.__setitem__, 'LIST', [1.,'TWO'])
        self.assertRaises(VicarError, vic.__setitem__, 'LIST', ['ONE', set()])
        self.assertRaises(VicarError, vic.__setitem__, 'LIST', [set(), 2])

        self.assertRaises(TypeError, vic.__setitem__, set(), 7)
        self.assertRaises(TypeError, vic.__setitem__, 3.14, 'pi')
        self.assertRaises(VicarError, vic.__setitem__, 0, set())

        self.assertRaises(VicarError, vic.__setitem__, 'ORG', 'whatever')
        self.assertRaises(VicarError, vic.__setitem__, 'NL', 'whatever')
        self.assertRaises(VicarError, vic.__setitem__, 'NL', -1)

        # __setitem__, value_str
        vic['A'] = 1
        self.assertEqual(vic.value_str('A'), '1')
        vic['A'] = (1, '%+03d')
        self.assertEqual(vic.value_str('A'), '+01')
        vic['A'] = 3
        self.assertEqual(vic.value_str('A'), '+03')         # format preserved
        vic['A'] = (1, '%+03d', 3)
        self.assertEqual(vic.value_str('A'), '+01   ')
        vic['A'] = 4
        self.assertEqual(vic.value_str('A'), '+04   ')
        vic['A'] = 4.
        self.assertEqual(vic.value_str('A'), '4.   ')       # format replaced
        vic['A'] = 4
        self.assertEqual(vic.value_str('A'), '4   ')
        vic['A'] = (1, '%+03d', 3, 0)
        self.assertEqual(vic.value_str('A'), '   +01')
        vic['A'] = 4
        self.assertEqual(vic.value_str('A'), '   +04')
        vic['A'] = (1, 3, 0)
        self.assertEqual(vic.value_str('A'), '   1')
        vic['A'] = 4
        self.assertEqual(vic.value_str('A'), '   4')
        vic['A'] = (1, 3)
        self.assertEqual(vic.value_str('A'), '1   ')
        vic['A'] = 4
        self.assertEqual(vic.value_str('A'), '4   ')

        vic['B'] = 1.234
        self.assertEqual(vic.value_str('B'), '1.234')
        vic['B'] = (1.234, '%#+.4f')
        self.assertEqual(vic.value_str('B'), '+1.2340')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '+1.2346')     # format preserved
        vic['B'] = (1.234, '%#.0f')
        self.assertEqual(vic.value_str('B'), '1.')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '1.')
        vic['B'] = (1.234, '%#+.4f', 3)
        self.assertEqual(vic.value_str('B'), '+1.2340   ')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '+1.2346   ')
        vic['B'] = 1
        self.assertEqual(vic.value_str('B'), '1   ')        # format replaced
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '1.23456   ')  # format replaced
        vic['B'] = (1.234, '%#+.4f', 3, 0)
        self.assertEqual(vic.value_str('B'), '   +1.2340')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '   +1.2346')
        vic['B'] = (1.234, 3, 0)
        self.assertEqual(vic.value_str('B'), '   1.234')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '   1.23456')
        vic['B'] = (1.234, 3)
        self.assertEqual(vic.value_str('B'), '1.234   ')
        vic['B'] = 1.23456
        self.assertEqual(vic.value_str('B'), '1.23456   ')

        vic['C'] = 1.234999993535
        self.assertEqual(vic.value_str('C'), '1.235')
        vic['C'] = -1.23500000123
        self.assertEqual(vic.value_str('C'), '-1.235')
        vic['C'] = 1.000001999999
        self.assertEqual(vic.value_str('C'), '1.')
        vic['C'] = -1.9999900000
        self.assertEqual(vic.value_str('C'), '-2.')
        vic['C'] = 9.9999900000
        self.assertEqual(vic.value_str('C'), '10.')
        vic['C'] = 1.899999050000044
        self.assertEqual(vic.value_str('C'), '1.9')
        vic['C'] = -1.02999998434
        self.assertEqual(vic.value_str('C'), '-1.03')

        vic['D'] = -1.234999993535e-12
        self.assertEqual(vic.value_str('D'), '-1.235E-12')
        vic['D'] = 1.23500000123e-12
        self.assertEqual(vic.value_str('D'), '1.235E-12')
        vic['D'] = -1.000001999999e-12
        self.assertEqual(vic.value_str('D'), '-1.E-12')
        vic['D'] = 1.9999900000e-12
        self.assertEqual(vic.value_str('D'), '2.E-12')
        vic['D'] = -9.9999900000e-12
        self.assertEqual(vic.value_str('D'), '-1.E-11')
        vic['D'] = 1.899999050000044e-12
        self.assertEqual(vic.value_str('D'), '1.9E-12')
        vic['D'] = -1.02999998434e-12
        self.assertEqual(vic.value_str('D'), '-1.03E-12')

        vic['E'] = 7.
        self.assertEqual(vic.value_str('E'), '7.')

        vic['F'] = [1.234, -1.234999993535e-12]
        self.assertEqual(vic.value_str('F'), '(1.234,-1.235E-12)')
        vic['F'] = [(1.234, '%#+.4f'), -1.23500000123, -1.000001999999e-12]
        self.assertEqual(vic.value_str('F'), '(+1.2340,-1.235,-1.E-12)')
        vic['F'] = [(1.234, 1, 0), (7., 1), -1.02999998434e-12]
        self.assertEqual(vic.value_str('F'), '( 1.234,7. ,-1.03E-12)')
        vic['F'] = ([(1.234, 1, 0), (7., 1), -1.02999998434e-12], 2, 3)
        self.assertEqual(vic.value_str('F'), '  ( 1.234,7. ,-1.03E-12)   ')

        vic['G'] = 'xyz'
        self.assertEqual(vic.value_str('G'), "'xyz'")
        vic['G'] = ('xyz', 3)
        self.assertEqual(vic.value_str('G'),"'xyz'   ")
        vic['G'] = 'abc'
        self.assertEqual(vic.value_str('G'),"'abc'   ")
        vic['G'] = ('abc', 3, 1)
        self.assertEqual(vic.value_str('G'),"   'abc' ")
        vic['G'] = 'xyz'
        self.assertEqual(vic.value_str('G'),"   'xyz' ")

        vic['H'] = [1,2,3]
        self.assertEqual(vic.value_str('H'), "(1,2,3)")
        vic['H'] = ([1,2,3], 2, 1)
        self.assertEqual(vic.value_str('H'), "  (1,2,3) ")
        vic['H'] = [4,5,6,7]
        self.assertEqual(vic.value_str('H'), "  (4,5,6,7) ")
        vic['H'] = [(1,1,1), (2,2,2), (3,3,3)]
        self.assertEqual(vic.value_str('H'), "( 1 ,  2  ,   3   )")
        vic['H'] = [4,5,6,7]
        self.assertEqual(vic.value_str('H'), "(4,5,6,7)")
        vic['H'] = ([(1,1,1), (2,2,2), (3,3,3)], 2, 1)
        self.assertEqual(vic.value_str('H'), "  ( 1 ,  2  ,   3   ) ")
        vic['H'] = [4,5,6,7]
        self.assertEqual(vic.value_str('H'), "  (4,5,6,7) ")

        vic['ORG+'] = 'not BSQ or BIL or BIP'   # not a VicarError
        self.assertEqual(vic.value_str(('ORG',-1)), "'not BSQ or BIL or BIP'")

        # copy
        vic2 = vic.copy()
        self.assertEqual(vic, vic2)
        vic2['E'] = 9.
        self.assertEqual(vic['E'], 7.)
        vic2['NEW_ITEM'] = 'This is a new item'
        self.assertEqual(len(vic), len(vic2) - 1)

        # reorder
        vic.reorder('', 'A', 'B', 'C', 'F')
        self.assertEqual(vic.names()[:6], ['LBLSIZE', 'A', 'B', 'C', 'F', 'FORMAT'])
        self.assertRaises(KeyError, vic.reorder, 'A', 'UNK', 'C', 'F')
        self.assertRaises(ValueError, vic.reorder, 'A', 'F', 'C', 'F')

        vic.reorder('FORMAT', 'F', 'G')
        self.assertEqual(vic.names()[:7], ['LBLSIZE', 'A', 'B', 'C', 'FORMAT', 'F', 'G'])

        # _set_n321, _set_nbls, _n123_from_nbls, _nbls_from_n123
        vic['ORG'] = 'BSQ'
        vic._set_n321(300, 200, 100)
        self.assertEqual((vic['N3'], vic['N2'], vic['N1']), (300, 200, 100))
        self.assertEqual((vic['NB'], vic['NL'], vic['NS']), (300, 200, 100))
        vic._set_nbls(101, 201, 301)
        self.assertEqual((vic['NB'], vic['NL'], vic['NS']), (101, 201, 301))
        self.assertEqual((vic['N3'], vic['N2'], vic['N1']), (101, 201, 301))

        vic['ORG'] = 'BIL'
        vic._set_n321(300, 200, 100)
        self.assertEqual((vic['N3'], vic['N2'], vic['N1']), (300, 200, 100))
        self.assertEqual((vic['NL'], vic['NB'], vic['NS']), (300, 200, 100))
        vic._set_nbls(101, 201, 301)
        self.assertEqual((vic['NB'], vic['NL'], vic['NS']), (101, 201, 301))
        self.assertEqual((vic['N2'], vic['N3'], vic['N1']), (101, 201, 301))

        vic['ORG'] = 'BIP'
        vic._set_n321(300, 200, 100)
        self.assertEqual((vic['N3'], vic['N2'], vic['N1']), (300, 200, 100))
        self.assertEqual((vic['NL'], vic['NS'], vic['NB']), (300, 200, 100))
        vic._set_nbls(101, 201, 301)
        self.assertEqual((vic['NB'], vic['NL'], vic['NS']), (101, 201, 301))
        self.assertEqual((vic['N1'], vic['N3'], vic['N2']), (101, 201, 301))

        # Reading image file C0532836239R.IMG

        filepath = test_dir / 'C0532836239R.IMG'

        vic = VicarLabel(filepath)

        self.assertEqual(vic.filepath, filepath)
        self.assertEqual(vic['NL'], 800)
        self.assertEqual(vic['TRUTH_WINDOW'], [801,801,96,96])
        self.assertEqual(vic['COMPRESSION_RATIO'], 9.64155)
        self.assertEqual(vic['SOLRANGE'], 7.43341E+08)
        self.assertEqual(vic['TARGET'], 'EUROPA')

        test = VicarLabel.read_label(filepath)
        self.assertEqual(vic, VicarLabel(test))

        f = filepath.open('rb')
        test2 = VicarLabel.read_label(f)
        self.assertEqual(test2, test)
        f.close()

        (test3, extra) = VicarLabel.read_label(filepath, _extra=True)
        self.assertEqual(test3, test)
        size = vic['LBLSIZE'] + (vic['NL'] + vic['NLB']) * vic['RECSIZE']
        self.assertEqual(size + len(extra), os.path.getsize(filepath))
        self.assertTrue(all([c == 0 for c in extra]))

        # New __setitem__ features 2/28/24
        lbl = VicarLabel()
        lbl['NEW1', 0] = 1
        lbl['NEW2+', 0] = 2
        lbl['NEW3', 0] = 3
        lbl['NEW4+', 0] = 4
        lbl['NEW5'] = 5
        lbl['NEW5', 1] = 6
        self.assertEqual(lbl['NEW1'], 1)
        self.assertEqual(lbl['NEW2'], 2)
        self.assertEqual(lbl['NEW3'], 3)
        self.assertEqual(lbl['NEW4'], 4)
        self.assertEqual(lbl['NEW5+'], [5,6])

        self.assertRaises(KeyError, lbl.__setitem__, ('NEW6', 1), 7)
        self.assertRaises(KeyError, lbl.__setitem__, ('NEW6',-2), 8)

        # New features 2/29/24
        # append() using a dictionary
        lbl = VicarLabel()
        lbl.append({'A':1, 'B':'TWO', 'C':(3, '%03d', 1, 1)})
        self.assertEqual(lbl['A'], 1)
        self.assertEqual(lbl['B'], 'TWO')
        self.assertEqual(lbl['C'], 3)
        self.assertEqual(lbl._formats[-1], ('%03d', 0, 1, 1, []))

        # arg() with a value
        lbl = VicarLabel()
        lbl.append({'TASK':'FICOR77', 'FOO':1})
        lbl.append({'TASK':'RESLOC', 'BAR':2})
        lbl.append({'TASK':'GEOMA', 'FOO':3, 'BAR':4})

        self.assertEqual(lbl.arg('TASK', 'GEOMA'), len(lbl) - 3)
        self.assertEqual(lbl.arg('TASK', 'FICOR77'), len(lbl) - 7)
        self.assertEqual(lbl.arg(len(lbl) - 3, 'GEOMA'), len(lbl) - 3)
        self.assertRaises(ValueError, lbl.arg, 'TASK', 'whatever')
        self.assertRaises(ValueError, lbl.arg, len(lbl) - 3, 'RESLOC')

        # arg() with a new key
        self.assertEqual(lbl.arg(('FOO', 'TASK', 'FICOR77'), 1), len(lbl) - 6)
        self.assertEqual(lbl.arg(('BAR', 'TASK', 'RESLOC'), 2), len(lbl) - 4)
        self.assertEqual(lbl.arg(('FOO', 'TASK', 'GEOMA'), 3), len(lbl) - 2)
        self.assertEqual(lbl.arg(('BAR', 'TASK', 'GEOMA'), 4), len(lbl) - 1)

        self.assertEqual(lbl.arg(('FOO', 'TASK', 'FICOR77')), len(lbl) - 6)
        self.assertEqual(lbl.arg(('BAR', 'TASK', 'RESLOC')), len(lbl) - 4)
        self.assertEqual(lbl.arg(('FOO', 'TASK', 'GEOMA')), len(lbl) - 2)
        self.assertEqual(lbl.arg(('BAR', 'TASK', 'GEOMA')), len(lbl) - 1)

        self.assertRaises(ValueError, lbl.arg, ('FOO', 'TASK', 'GEOMA'), 99)
        self.assertRaises(KeyError, lbl.arg, ('FOO', 'TASK', 'RESLOC'))

        # __getitem__() with a new key
        self.assertEqual(lbl['FOO', 'TASK'], 1)
        self.assertEqual(lbl['FOO', 'TASK', 'FICOR77'], 1)
        self.assertEqual(lbl['BAR', 'TASK', 'RESLOC'], 2)
        self.assertEqual(lbl['FOO', 'TASK', 'GEOMA'], 3)
        self.assertEqual(lbl['BAR', 'TASK', 'GEOMA'], 4)

        self.assertRaises(KeyError, lbl.__getitem__, ('BAR', 'TASK'))
        self.assertRaises(KeyError, lbl.__getitem__, ('BAR', 'TASK', 'FICOR77'))
        self.assertRaises(KeyError, lbl.__getitem__, ('FOO', 'TASK', 'RESLOC'))

        self.assertRaises(KeyError, lbl.__getitem__, ('FOO', 'TASKX'))
        self.assertRaises(KeyError, lbl.__getitem__, ('FOO', 'TASKX', 'whatever'))
        self.assertRaises(ValueError, lbl.__getitem__, ('FOO', 'TASK', 'RESSAR77'))

        # get() with a new key
        test = VicarLabel()
        self.assertEqual(test.get('FOO', 7), 7)
        self.assertEqual(test.get('FOO+', 7), [7])
        self.assertEqual(test.get(set(), 7), 7)

        # __setitem__() with a new key, existing
        lbl['FOO', 'TASK', 'FICOR77'] = 77
        self.assertEqual(lbl['FOO', 'TASK', 'FICOR77'], 77)
        self.assertEqual(lbl[len(lbl) - 6], 77)

        lbl['FOO', 'TASK', 'GEOMA'] = 88
        self.assertEqual(lbl['FOO', 'TASK', 'GEOMA'], 88)
        self.assertEqual(lbl[len(lbl) - 2], 88)

        lbl['FOO+', 'TASK', 'GEOMA'] = 33
        self.assertEqual(lbl['FOO+', 'TASK', 'GEOMA'], [88,33])
        del lbl['FOO', 'TASK', 'GEOMA']
        self.assertEqual(lbl['FOO+', 'TASK', 'GEOMA'], [33])

        # __contains__() with an integer key
        self.assertTrue(-len(lbl) in lbl)
        self.assertTrue(len(lbl) - 1 in lbl)
        self.assertFalse(-len(lbl) - 1 in lbl)
        self.assertFalse(len(lbl) in lbl)

        # __contains__() with a new key
        self.assertTrue('FOO' in lbl)
        self.assertTrue(('FOO', 'TASK') in lbl)
        self.assertTrue(('FOO', 'TASK', 'FICOR77') in lbl)
        self.assertFalse(('FOO', 'TASK', 'RESLOC') in lbl)
        self.assertTrue(('FOO', 'TASK', 'GEOMA') in lbl)

        self.assertFalse(('BAR', 'TASK') in lbl)
        self.assertFalse(('BAR', 'TASK', 'FICOR77') in lbl)
        self.assertTrue(('BAR', 'TASK', 'RESLOC') in lbl)
        self.assertTrue(('BAR', 'TASK', 'GEOMA') in lbl)

        # __setitem__() with a new key, not existing
        indx = lbl.arg('TASK', 'RESLOC')
        lbl['BAR', 'TASK', 'FICOR77'] = 99  # insertion
        self.assertEqual(lbl.arg(('BAR', 'TASK', 'FICOR77')), indx)
        self.assertTrue(('BAR', 'TASK', 'FICOR77') in lbl)
        names = list(lbl.names())
        self.assertEqual(names[-8:], ['TASK', 'FOO', 'BAR', 'TASK', 'BAR',
                                      'TASK', 'BAR', 'FOO'])

        indx = len(lbl)
        lbl['NEW', 'TASK', 'GEOMA'] = 99
        self.assertEqual(lbl.arg(('NEW', 'TASK', 'GEOMA')), indx)
        self.assertTrue(('NEW', 'TASK', 'GEOMA') in lbl)
        names = list(lbl.names())
        self.assertEqual(names[-9:], ['TASK', 'FOO', 'BAR', 'TASK', 'BAR',
                                      'TASK', 'BAR', 'FOO', 'NEW'])

        # __delitem__() with new key
        del lbl['FOO', 'TASK', 'GEOMA']
        names = list(lbl.names())
        self.assertEqual(names[-3:], ['TASK', 'BAR', 'NEW'])

        # __delitem__(), insert()
        test = VicarLabel()
        ltest = len(test)
        test.insert('FOO=1', 1)
        test.append('FOO=2')
        test.append('FOO=3')
        self.assertEqual(len(test), ltest+3)
        del test['FOO+',1]
        self.assertEqual(len(test), ltest+1)
        self.assertEqual(test[1], 1)

        self.assertRaises(VicarError, test.insert, 'BAR=7', 0)
        self.assertRaises(VicarError, test.__setitem__, 'LBLSIZE', [1,2])
        self.assertRaises(VicarError, test.__setitem__, 'LBLSIZE', 77.)
        self.assertRaises(VicarError, test.__setitem__, 0, [1,2])
        self.assertRaises(VicarError, test.__setitem__, 0, 77.)

##########################################################################################



##########################################################################################
# vicar/_DEFINITIONS.py
##########################################################################################
"""Definitions of global variables defining attributes of the VICAR standard.

Source: https://github.com/SETI/rms-vicar/blob/main/vicar/_DEFINITIONS.py
"""

import sys

# Fixed space between "LBLSIZE=" and the next parameter name
_LBLSIZE_WIDTH = 16

# [(sys.byteorder,sys.platform)] -> HOST
_HOST_DICT = {('big'   , 'sunos3'): 'SUN-3',
              ('big'   , 'sunos4'): 'SUN-4',
              ('big'   , 'sunos5'): 'SUN-SOLR',
              ('little', 'sunos5'): 'X86-LINUX',
              ('big'   , 'darwin'): 'MAC-OSX',
              ('little', 'darwin'): 'MAC-OSX',
              ('little', 'linux2'): 'X86-LINUX',
              ('little', 'linux3'): 'X86-LINUX',
              ('little', 'linux' ): 'X86-LINUX',
              ('little', 'win32' ): 'WIN-XP'     }

# Infer the _HOST dictionary key for this platform
try:
    _HOST = _HOST_DICT[(sys.byteorder, sys.platform)]
except KeyError:                # pragma: no cover
    if sys.platform.startswith('linux'):
        _HOST = 'X86-LINUX'     # could be "linux4" I guess
    else:
        _HOST = sys.platform.upper()

# [FORMAT] -> dtype
_DTYPE_FROM_FORMAT = {'BYTE': 'u1',
                      'HALF': 'i2',
                      'FULL': 'i4',
                      'REAL': 'f4',
                      'DOUB': 'f8',
                      'COMP': 'c8',
                      'WORD': 'i2',
                      'LONG': 'i4',
                      'COMPLEX': 'c8'}

# [dtype.kind + str(dtype.itemsize)] -> (FORMAT, isint)
_FORMAT_FROM_DTYPE = {'u1': ('BYTE', True ),
                      'i2': ('HALF', True ),
                      'i4': ('FULL', True ),
                      'f4': ('REAL', False),
                      'f8': ('DOUB', False),
                      'c8': ('COMP', False)}

# [sys.byteorder] -> INTFMT, REALFMT
_INTFMT_DICT  = {'little': 'LOW'  , 'big': 'HIGH'}
_REALFMT_DICT = {'little': 'RIEEE', 'big': 'IEEE'}

# Required keywords, default values
_REQUIRED = [('LBLSIZE' , 0,     ),
             ('FORMAT'  , 'BYTE' ),     # Guess
             ('TYPE'    , 'IMAGE'),     # Guess
             ('BUFSIZ'  , 20480  ),     # Always ignored
             ('DIM'     , 3      ),     # Always
             ('EOL'     , 0      ),
             ('RECSIZE' , 0      ),
             ('ORG'     , 'BSQ'  ),
             ('NL'      , 0      ),
             ('NS'      , 0      ),
             ('NB'      , 0      ),
             ('N1'      , 0      ),
             ('N2'      , 0      ),
             ('N3'      , 0      ),
             ('N4'      , 0      ),     # Always
             ('NBB'     , 0      ),
             ('NLB'     , 0      ),
             ('HOST'    , _HOST  ),
             ('INTFMT'  , _INTFMT_DICT [sys.byteorder]),
             ('REALFMT' , _REALFMT_DICT[sys.byteorder]),
             ('BHOST'   , _HOST  ),
             ('BINTFMT' , _INTFMT_DICT [sys.byteorder]),
             ('BREALFMT', _REALFMT_DICT[sys.byteorder]),
             ('BLTYPE'  , ''),]
_REQUIRED_NAMES = set([t[0] for t in _REQUIRED])

# Keywords that the user cannot modify
_IMMUTABLE = set(['LBLSIZE' ,
                  'FORMAT'  ,
                  'TYPE'    ,
                  'DIM'     ,
                  'EOL'     ,
                  'RECSIZE' ,
                  'ORG'     ,
                  'NL'      ,
                  'NS'      ,
                  'NB'      ,
                  'N1'      ,
                  'N2'      ,
                  'N3'      ,
                  'N4'      ,
                  'NBB'     ,
                  'NLB'     ,
                  'INTFMT'  ,
                  'REALFMT' ,
                  'BINTFMT' ,
                  'BREALFMT'])

# Keywords with enumerated values
_ENUMERATED_VALUES = {
    'FORMAT'  : {'BYTE', 'HALF', 'FULL', 'REAL', 'DOUB', 'COMP',
                 'WORD', 'LONG', 'COMPLEX'},
    'ORG'     : {'BSQ', 'BIL', 'BIP'},
    'INTFMT'  : {'HIGH', 'LOW'},
    'REALFMT' : {'IEEE', 'RIEEE', 'VAX'},
    'BINTFMT' : {'HIGH', 'LOW'},
    'BREALFMT': {'IEEE', 'RIEEE', 'VAX'},
    'DIM'     : {3},
    'EOL'     : {0, 1},
    'N4'      : {0},
}

# Keywords that must be positive ints
_REQUIRED_INTS = {'LBLSIZE', 'RECSIZE', 'NL', 'NS', 'NB', 'N1', 'N2', 'N3', 'NBB', 'NLB'}

##########################################################################################



##########################################################################################
# vicar/_LABEL_GRAMMAR.py
##########################################################################################
"""\
Definition of _LABEL_GRAMMAR, a `pyparsing` grammar supporting the VICAR format standard,
used internally. The full standard is defined here:

    https://pds-rings.seti.org/help/VICAR_file_fmt.pdf

_LABEL_GRAMMAR.parse_string(text) receives the contents of VICAR label and returns a list
of tuples. Each tuple defines the name and value of a label parameter. When the formatting
of that "name=value" pair is non-standard, the tuple provides additional information
indicating how the name and value were formatted. The tuple contains up to six items:

   (`name`, `value`, `format`, `after_name`, `before_value`, `after_value`)

where:

    * `name`: VICAR parameter name.
    * `value`: Value of the parameter, represented by an int, float, string, or list.
    * `format`: Optional format by which this value is represented in the label; included
      only for ints and floats and only if the format is different from the representation
      produced by str(value).
    * `name_blanks`:  number of blank characters after the name and before the equal sign;
      included only if this number is nonzero.
    * `value_blanks`: number of blank characters after the equal sign and before the
      value; included if this number is nonzero or if name_blanks is nonzero.
    * `sep_blanks`: number of blank characters after this value and before either the next
      value or the end of the label; included if this value does not equal 2 or if either
      of the previous values is nonzero.

In other words, the possible tuples returned are these:

    * (`name`, `value`)
    * (`name`, `value`, `format`)
    * (`name`, `value`, `format`, `sep_blanks`)
    * (`name`, `value`, `format`, `value_blanks`, `sep_blanks`)
    * (`name`, `value`, `format`, `name_blanks`, `value_blanks`, `sep_blanks`)
    * (`name`, `value`, `sep_blanks`)
    * (`name`, `value`, `value_blanks`, `sep_blanks`)
    * (`name`, `value`, `name_blanks`, `value_blanks`, `sep_blanks`)

In summary:

    * The first two items in the tuple are always `name` and `value`.
    * The third item, if present and a string, is `format`.
    * Subsequent items are integers that define the number of blank characters surrounding
      the equal sign and following the value's end; these are included only if one or more
      of their values are non-standard.

If the label value is a list, then each value in the parsed list is an int, float, or
tuple containing up to four values:

  * `value`: The value of this item as an in int, float, or string.
  * `format`: Optional format by which this value is represented in the label; included
    only for ints and floats and only if the format is different from the representation
    produced by str(value).
  * `blanks_before`: Number of blanks before the item and after the left parenthesis or
    comma; included only if this number is nonzero.
  * `blanks_after`: Number of blanks after the value and before the next comma or the
    right parenthesis; included if this number is nonzero or if blanks_before is nonzero.

In other words, the possible items in a parsed list are these:

    * `value`
    * (`value`, `format`)
    * (`value`, `format`, `blanks_after`)
    * (`value`, `format`, `blanks_before`, `blanks_after`)
    * (`value`, `blanks_after`)
    * (`value`, `blanks_before`, `blanks_after`)

Most users will only care about the name and value, but the remaining information in each
tuple is included in case the user wants to write out a label string that preserves the
formatting of the original source.

Source: https://github.com/SETI/rms-vicar/blob/main/vicar/_LABEL_GRAMMAR.py
"""
##########################################################################################

from pyparsing import (alphanums,
                       alphas,
                       CharsNotIn,
                       Combine,
                       FollowedBy,
                       Literal,
                       nums,
                       oneOf,
                       OneOrMore,
                       Optional,
                       ParserElement,
                       StringEnd,
                       Suppress,
                       White,
                       Word,
                       ZeroOrMore)

ParserElement.set_default_whitespace_chars(' ')

_BREAK = ((ZeroOrMore(White(' ')) + (FollowedBy(oneOf(', )'))
                                     | (Suppress(ZeroOrMore('\0')) + StringEnd())))
          | OneOrMore(White(' \t\n\r')))

_OPT_WHITE = ZeroOrMore(White(' '))

############################################
# _INTEGER
############################################

def _int_info(token):
    rstripped = token.rstrip()
    len_rstripped = len(rstripped)
    after = len(token) - len_rstripped

    stripped = rstripped.lstrip()
    len_stripped = len(stripped)
    before = len_rstripped - len_stripped

    value = int(stripped)

    plus = '+' if stripped[0] == '+' else ''
    unsigned = stripped[1:] if stripped[0] in '+-' else stripped
    zero = unsigned[0] == '0' and len(unsigned) > 1
    if zero:
        fmt = '%' + plus + '0' + str(len_stripped) + 'd'
    elif plus:
        fmt = '%+d'
    else:
        fmt = ''

    return (value, fmt, before, after)

_INT = Word(nums)
_OPT_INT = Optional(_INT)
_OPT_SIGN = Optional(oneOf('+ -'))

_INTEGER = Combine(_OPT_WHITE + _OPT_SIGN + _INT + _BREAK)
_INTEGER.set_name('_INTEGER')
_INTEGER.set_parse_action(lambda s,l,t: _int_info(t[0]))

############################################
# _FLOAT
############################################

def _float_info(token):
    rstripped = token.rstrip()
    len_rstripped = len(rstripped)
    after = len(token) - len_rstripped

    stripped = rstripped.lstrip()
    len_stripped = len(stripped)
    before = len_rstripped - len_stripped

    value = float(stripped)

    mantissa, e, expo = stripped.partition('e')
    plus = '+' if mantissa[0] == '+' else ''
    unsigned = mantissa[1:] if mantissa[0] in '+-' else mantissa
    (head, dot, tail) = unsigned.partition('.')

    if e:
        prec = len(head) + len(tail) - 1
        letter = 'E'
    else:
        prec = len(tail)
        letter = 'f'

    fmt = '%#' + plus + '.' + str(prec) + letter

    return (value, fmt, before, after)

_DOT = Literal('.')

_EXPO = Combine(Suppress(oneOf('e E d D')) + _OPT_SIGN + _INT)
_EXPO.set_parse_action(lambda s,l,t: 'e' + t[0])
_OPT_EXPO = Optional(_EXPO)

_FLOAT_W_INT = Combine(_OPT_WHITE + _OPT_SIGN + _INT + _DOT + _OPT_INT + _OPT_EXPO
                       + _BREAK)
_FLOAT_WO_INT = Combine(_OPT_WHITE + _OPT_SIGN + _DOT + _INT + _OPT_EXPO + _BREAK)
_FLOAT_WO_DOT = Combine(_OPT_WHITE + _OPT_SIGN + _INT + _EXPO + _BREAK)
_FLOAT = _FLOAT_W_INT | _FLOAT_WO_INT | _FLOAT_WO_DOT
_FLOAT.set_name('_FLOAT')
_FLOAT.set_parse_action(lambda s,l,t: _float_info(t[0]))

############################################
# _STRING
############################################

def _str_info(token):
    rstripped = token.rstrip()
    after = len(token) - len(rstripped)

    stripped = rstripped.lstrip()
    before = len(rstripped) - len(stripped)

    value = stripped[1:-1].replace("''", "'")
    return (value, '', before, after)

_QUOTE = Literal("'")
_QQUOTE = Literal("''")
_STRING = Combine(_OPT_WHITE + _QUOTE + ZeroOrMore(CharsNotIn("'") | _QQUOTE) + _QUOTE
                  + _BREAK)
_STRING.set_name('_STRING')
_STRING.set_parse_action(lambda s,l,t: _str_info(t[0]))

############################################
# _LIST
############################################

def _list_info(tokens):
    before = len(tokens[0])
    after = len(tokens[-1])

    return (tokens[1:-1], '', before, after)

_SCALAR = _INTEGER | _FLOAT | _STRING

_LPAREN = Combine(_OPT_WHITE + Suppress(Literal('(')))
_RPAREN = Combine(Suppress(Literal(')')) + _BREAK)
_COMMA = Suppress(Literal(','))

_LIST = _LPAREN + _SCALAR + ZeroOrMore(_COMMA + _SCALAR) + _RPAREN
_LIST.set_name('_LIST')
_LIST.set_parse_action(lambda s,l,t: _list_info(t))

############################################
# _VALUE
############################################

_VALUE = _SCALAR | _LIST

############################################
# _NAME
############################################

def _name_info(token):
    rstripped = token.rstrip()
    after = len(token) - len(rstripped)
    return (rstripped, after)

# Strict version:
# _ALPHAS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
# _ALNUMS = _ALPHAS + '0123456789_'
# _NAME = Combine(Word(_ALPHAS, _ALNUMS, max=32) + _OPT_WHITE)

# Loose version allows mixed case and more than 32 characters
_NAME = Combine(Word(alphas, alphanums + '_') + _OPT_WHITE)
_NAME.set_name('_NAME')
_NAME.set_parse_action(lambda s,l,t: _name_info(t[0]))

############################################
# _STATEMENT
############################################

def _statement_info(tokens):
    (name, blanks0) = tokens[0]
    (value, fmt, blanks1, blanks2) = tokens[1]

    # Strip unnecessary formatting info from the items in a list
    if isinstance(value, list):
        new_value = []
        for (item_value, item_fmt, before, after) in value:
            result = [item_value]
            if item_fmt:
                result.append(item_fmt)
            if before:
                result += [before, after]
            elif after != 0:
                result += [after]

            if len(result) == 1:
                new_value.append(result[0])
            else:
                new_value.append(tuple(result))

        value = new_value

    # Construct the result as a new tuple
    result = [name, value]
    if fmt:
        result.append(fmt)

    # Append up to three counts of blank values
    if blanks0:
        result += [blanks0, blanks1, blanks2]
    elif blanks1:
        result += [blanks1, blanks2]
    elif blanks2 != 2:
        result += [blanks2]

    return tuple(result)

_EQUAL = Suppress(Literal('='))
_STATEMENT = _NAME + _EQUAL + _VALUE
_STATEMENT.set_name('_STATEMENT')
_STATEMENT.set_parse_action(lambda s,l,t: _statement_info(t))

############################################
# _LABEL_GRAMMAR
############################################

_LABEL_GRAMMAR = OneOrMore(_STATEMENT) + StringEnd()
_LABEL_GRAMMAR.set_name('_LABEL_GRAMMAR')

##########################################################################################



##########################################################################################
# vicar/__init__.py
##########################################################################################
"""
VICAR File Support Module,
PDS Ring-Moon Systems Node, SETI Institute

**Introduction**

`vicar` is a Python module that supports reading and writing of JPL’s VICAR file format.
It supports the definition of the VICAR file format as found here:

    https://pds-rings.seti.org/help/VICAR_file_fmt.pdf

**Getting Started**

The `vicar` modules provies these classes:

    * `VicarLabel`: Class for reading, writing, and parsing of VICAR labels.
    * `VicarImage`: Class for handling VICAR image (and other) data files.
    * `VicarError`: Extension of class ValueError to contain exceptions.

To read a VICAR image file::

    import vicar
    vic = vicar.VicarImage("path/to/file")

The returned `VicarImage` object "vic" has these attributes:

    * `vic.array`: The 3-D data array converted to native format.
    * `vic.array2d`: Same as above, but with leading dimension (typically, bands)
      stripped.
    * `vic.prefix`: The array prefix bytes as a 3-D array of unsigned bytes.
    * `vic.prefix2d`: Same as above, but with the leading dimension stripped.
    * `vic.binheader`: The binary header as a bytes object; use vic.binheader_array() to
      extract information.
    * `vic.label`: The internal VicarLabel object that manages the VICAR label
      information, if direct access is needed.

VICAR parameter values can be extracted from the label using dictionary-like syntax::

    len(vic)                # the number of parameters in the VICAR label.
    vic['LBLSIZE']          # the value of the LBLSIZE parameter (an integer).
    vic[0]                  # the value of the first parameter.
    vic[-1]                 # the value of the last parameter.
    vic['LBLSIZE',-1]       # the value of the last occurrence of the LBLSIZE parameter.
    vic.get(('LBLSIZE',2), 99)
                            # the value of the third occurrence of the LBLSIZE
                            # parameter, or 99 if there are fewer than 3 occurrences.
    vic.arg('LBLSIZE')      # the numeric index of "LBLSIZE" among the VICAR parameters.

You can also use dictionary-like syntax to modify and insert header values::

    vic['SOLDIST'] = 1.e9   # set SOLDICT to this value.
    del vic['SOLDIST',0]    # Remove the first occurrence of SOLDIST from the label.
    vic['LBLSIZE+'] = 2000  # insert a new LBLSIZE parameter instead of modifying an
                            # existing one.

Note that certain required VICAR parameters contain structural information about the
file; these cannot generally be modified directly.

Numerous methods are available to iterate over the VICAR label parameters::

    for (name,value) in vic.items(): ...
    for key in vic.keys(): ...
    for name in vic.names(): ...
    for value in vic.values(): ...

Iterators can take a regular expression as input to restrict the items returned::

    for value in vic.values(r'LAB\\d\\d'): ...

Use `str(vic)` to get the VICAR label content represented as a string.

Here are the steps to create and write a VICAR image file::

    vic = VicarImage()
    vic.array = array
    vic.prefix = prefix
    vic.binheader = binheader
    vic['NOTES'] = ['add as many more VICAR parameters', 'as you wish']
    vic.write_file("path/to/file")
"""

__all__ = ['VicarError', 'VicarImage', 'VicarLabel']

from vicar.vicarlabel import VicarLabel, VicarError
from vicar.vicarimage import VicarImage

try:
    from ._version import __version__
except ImportError:
    __version__ = 'Version unspecified'

##########################################################################################



##########################################################################################
# vicar/vicarimage.py
##########################################################################################
"""Class to support accessing, reading, and modifying VICAR image files."""

import numpy as np
import pathlib
import sys
import vax
import warnings

try:
    from _version import __version__
except ImportError:
    __version__ = 'Version unspecified'

from vicar.vicarlabel import VicarLabel, VicarError
from vicar._DEFINITIONS import _DTYPE_FROM_FORMAT, _FORMAT_FROM_DTYPE, _HOST, _IMMUTABLE


class VicarImage():
    """Constructor for a VicarImage.

    This class defines the contents of a VICAR data file. It supports methods for reading
    and writing files and for accessing header information.

    **Properties**:

        * `array`: The 3-D data array converted to native format.
        * `array3d`: Same as above.
        * `array2d`: Same as above, but with leading dimension (typically, bands)
          stripped.
        * `prefix`: The array prefix bytes as a 3-D array of unsigned bytes.
        * `prefix3d`: Same as above.
        * `prefix2d`: Same as above, but with the leading dimension stripped.
        * `binheader`: The binary header as a bytes object; use vic.binheader_array() to
          extract information.
        * `label`: The internal object that manages the VICAR label information, available
          if direct access to it is needed.

    **Core Methods:**

        * `arg`: The index of a parameter name within this label.
        * `args`: Iterator over the numeric indices of the parameters in a label.
        * `binheader_array`: Interpret the content of a binary header.
        * `copy`: A copy of this VicarImage.
        * `deepcopy`: An independent (deep) copy of this VicarImage.
        * `from_array`: Construct a VicarImage object for a NumPy array.
        * `from_file`: Construct a VicarImage object from the content of a VICAR data
          file.
        * `get`: Retrieve a label parameter value or return a default.
        * `items`: Iterator over the (name, value) tuples in the label.
        * `keys`: Iterator over the parameter names in the label as unique keys.
        * `names`: Iterator over the parameter names in the label.
        * `values`: Iterator over the parameter values in the label.
        * `write_file`: Write this object as a VICAR data file.

    **Python Syntax Support Methods:**

        * `__contains__`: Enables "`name in image`" syntax for checking a name in the
          label.
        * `__delitem__`: Enables "`del image[name]`" syntax to remove a label parameter
          name.
        * `__eq__`: Enables "`a == b`', the test of whether two image objects are equal.
        * `__getitem__`: Enables "`image[name]`" dictionary-like syntax to get the value
          of a label parameter.
        * `__iter__`: Enables "`for key in image:`" syntax to iterate over the label
          parameter keys.
        * `__len__`: Enables "`len(image)`", the number of parameters in the image's
          label.
        * `__repr__`: Enables "`repr(image)`", similar to the "`str(label)`", but with the
          class name included.
        * `__setitem__`: Enables "`image[name] = value`" dictionary-like syntax to set the
          value of a label parameter.
        * `__str__`: Enables "`str(image)`", returning a string representing the content
          of a label.

    **Notes About Dictionary Keys:**

        When using dictionary-like syntax to reference a parameter in a VICAR label, a
        rich set of options are available. For example, if `image` is a VicarImage object,
        then:

            * `image[n]` where `n` is an integer refers to the "nth" parameter in the
              label. `n` can be positive or negative.
            * `image[name]` where `name` is a string refers to the first occurrence in
              the label of a parameter with this name.
            * `image[name,n]` refers to the "nth" occurrence in the label of a parameter
              with this name. `n` can be positive or negative.
            * `image[name, after]` where both items are strings refers to the first
              occurrence of parameter `name` after the first occurrence of parameter
              `after` and before the second occurrence of `after`.
            * `image[name, after, value]` refers to the first occurrence of parameter
              `name` after the first location where `after` equals `value` and before the
              next occurrence of `after`.

        The last two options make it easy to reference a VICAR label parameter that is
        repeated. For example, "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies
        the occurrence of `DAT_TIM` applicable to `TASK='COPY'` when there might be other
        `TASK` sections of the image label containing other values of `DAT_TIM`.

        Append a "+" to `name` to expand upon the function's behavior. With "get"
        operations, a list is returned identifying all of the occurrences of the selected
        name rather than just the first or "nth". With "set" operations, a new occurrence
        of `name` is inserted into the label even if the a parameter of the given name is
        already present.
    """

    def __init__(self, source=None, array=None, *, prefix=None, binheader=None,
                 strict=True):
        """Constructor for a VicarImage object.

        Parameters:
            source (pathlib.Path, str, or VicarLabel, optional):
                Source for the VicarImage based on a file path or VicarLabel; if not
                specified (equvalent to source=None), a minimal label is created.
            array (array-like, optional):
                Optional data array for this object. If the source is a file path,
                this array will override that in the file.
            prefix (array-like, optional):
                Optional prefix bytes for this object. If the source is a file
                path, this value will override that in the file. To remove the
                prefix array found in the file, use prefix=[].
            binheader (array-like or bytes, optional):
                Optional binary header for this data file. If the source is a
                file path, this value will override that in the file. To remove
                the binheader found in the file, use binheader=b''.
            strict (bool, optional):
                True (the default) to require strict conformance to the VICAR standard;
                False for a looser version of the standard. If `source` is a VicarLabel,
                this input is ignored; the value already assigned to the `source` is used
                instead.
        """

        self._filepath = None

        if not source:
            self._label = VicarLabel([], strict=strict)
        elif isinstance(source, VicarLabel):
            self._label = source
        else:
            self._filepath = pathlib.Path(source)
            info = VicarImage._read_file(self.filepath, extraneous='ignore',
                                         strict=strict)
            (self._label, array1, prefix1, binheader1) = info

            if array is None:
                array = array1
            if prefix is None:
                prefix = prefix1
            if binheader is None:
                binheader = binheader1

        # Validate array, prefix, and binheader using setters
        self._array = None
        self._prefix = None
        self._binheader = None

        self.array = array
        self.prefix = prefix
        self.binheader = binheader

    ######################################################################################
    # Properties
    ######################################################################################

    @property
    def filepath(self):
        """The path to the associated file.

        Returns:
            pathlib.Path or None: The path to the associated file, if any.
        """

        return self._filepath

    @filepath.setter
    def filepath(self, value):
        """Set the path to the associated file.

        Parameters:
            value (pathlib.Path, str, or None):
                The path to the associated file; None to remove a file association.
        """

        if value is None:
            self._filepath = None
        else:
            self._filepath = pathlib.Path(value)

    @property
    def label(self):
        """The VicarLabel object.

        Returns:
            VicarLabel: The VicarLabel object.
        """

        return self._label

    @property
    def array2d(self):
        """The data array as 2-D.

        Returns:
            np.ndarray or None:
                The data array as 2-dimensional if it is present; otherwise, None.

        Raises:
            VicarError: If the array has more than two dimensions.
        """

        if self._array is None:
            return None
        if self._array.shape[0] != 1:
            raise VicarError(f'VICAR array shape {self._array.shape} is not 2-D')
        return self._array[0]

    @property
    def data_2d(self):
        """The data array as 2-D; DEPRECATED name.

        Returns:
            np.ndarray or None:
                The data array as 2-dimensional if it is present; otherwise, None.

        Raises:
            VicarError: If the array has more than two dimensions.
        """

        return self.array2d

    @property
    def array3d(self):
        """The data array as 3-D.

        Returns:
            np.ndarray or None:
                The data array as 3-dimensional if it is present; otherwise, None.
        """

        return self._array

    @property
    def data_3d(self):
        """The data array as 3-D; DEPRECATED name.

        Returns:
            np.ndarray or None:
                The data array as 3-dimensional if it is present; otherwise, None.
        """

        return self._array

    @property
    def array(self):
        """The data array.

        Returns:
            np.ndarray or None: The data array if it is present; otherwise, None.
        """

        return self._array

    @array.setter
    def array(self, value):
        """Set the data array.

        Parameters:
            value (array-like or None):
                The prefix array or None to remove the data array.

        Raises:
            VicarError:
                If the array shape is incompatible with an existing prefix or binary
                header.
        """

        if value is None:
            self._array = None
            return

        value = np.asarray(value)
        if value.ndim != 3:
            value = value.reshape((3-value.ndim) * (1,) + value.shape)  # reshape to 3-D
        VicarImage._check_array_vs_prefix(value, self._prefix)

        recsize = self._label['NBB'] + value.shape[-1] * value.itemsize
        nlb = self._label['NLB']
        if self._binheader is not None:
            width = len(bytes(self._binheader))
            if width % recsize != 0:
                raise VicarError(f'array shape {value.shape} is incompatible with binary '
                                 f'header width {width}')
            nlb = width // recsize

        # Tests passed
        self._array = value
        self._label['HOST'] = _HOST
        self._label['TYPE'] = 'IMAGE'
        (self._label['FORMAT'], isint) = VicarImage._format_isint(value)
        if isint:
            if value.itemsize > 1:
                self._label['INTFMT'] = VicarImage._intfmt(value)
        else:
            self._label['REALFMT'] = VicarImage._realfmt(value)

        self._label['RECSIZE'] = recsize
        self._label['NLB'] = nlb
        self._label._set_n321(*value.shape)

    @property
    def prefix2d(self):
        """The image prefix as a 2-D array.

        Returns:
            np.ndarray: The image prefix as a 2-dimensional array.

        Raises:
            VicarError: If the prefix array has more than two dimensions.
        """

        if self._prefix is None:
            return None
        if self._prefix.shape[0] != 1:
            raise VicarError(f'prefix bytes shape {self._prefix.shape} is not 2-D')
        return self._prefix[0]

    @property
    def prefix_2d(self):
        """The image prefix as a 2-D array; DEPRECATED name.

        Returns:
            np.ndarray: The image prefix as a 2-dimensional array.

        Raises:
            VicarError: If the prefix array has more than two dimensions.
        """

        return self.prefix2d

    @property
    def prefix3d(self):
        """The image prefix as a 3-D array.

        Returns:
            np.ndarray: The image prefix as a 3-dimensional array.
        """

        return self._prefix

    @property
    def prefix_3d(self):
        """The image prefix as a 3-D array; DEPRECATED name.

        Returns:
            np.ndarray: The image prefix as a 2-dimensional array.
        """

        return self._prefix

    @property
    def prefix(self):
        """The image prefix object.

        Returns:
            np.ndarray: The image prefix as a 2-dimensional array.
        """

        return self._prefix

    @prefix.setter
    def prefix(self, value):
        """Set the image prefix array.

        Parameters:
            value (array-like or None):
                The prefix array or None to remove the prefix array.

        Raises:
            VicarError:
                If the array shape is incompatible with an existing data array or binary
                header.
        """

        if value is None:
            nbb = 0
        else:
            value = np.asarray(value)
            if value.size == 0:
                nbb = 0
                value = None
            else:
                if value.ndim != 3:
                    value = value.reshape((3-value.ndim) * (1,) + value.shape)
                VicarImage._check_array_vs_prefix(self._array, value)
                nbb = value.shape[-1] * value.itemsize

        recsize = self['RECSIZE']
        nlb = self['NLB']
        if self._array is not None:
            recsize = self._array.shape[-1] * self._array.itemsize + nbb
            if self._binheader is not None:
                width = len(bytes(self._binheader))
                if width % recsize != 0:
                    raise VicarError(f'new RECSIZE={recsize} is incompatible with binary '
                                     f'header length {width}')
                nlb = width // recsize

        # Tests passed
        self._prefix = value
        self._label['NBB'] = nbb
        self._label['NLB'] = nlb
        self._label['RECSIZE'] = recsize

        if value is not None:
            self._label['HOST'] = _HOST
            (fmt, isint) = VicarImage._format_isint(value)

            # Prefix defines attributes left undefined by array
            if self._array is None:
                self._label['FORMAT'] = fmt

            if isint:
                if self._array is None or self._array.dtype.kind in 'fc':
                    if value.itemsize > 1:
                        self._label['INTFMT'] = VicarImage._intfmt(value)
            else:
                if self._array is None or self._array.dtype.kind in 'ui':
                    self._label['REALFMT'] = VicarImage._realfmt(value)

    @property
    def binheader(self):
        """The binary header as an array or bytes object.

        Returns:
            np.ndarray, bytes, or None: The binary header if present; otherwise, None.
        """

        return self._binheader

    @binheader.setter
    def binheader(self, value):
        """Set the binary header.

        Parameters:
            value (bytes, np.ndarray, or None):
                The binary header or None to remove the binary header.

        Raises:
            VicarError:
                If the header length is incompatible with the existing RECSIZE.
        """

        if value is None:
            self._binheader = None
            self._label['NLB'] = 0
            return

        nlb = self._label['NLB']
        width = len(bytes(value))
        if self._array is not None:
            recsize = self._label['RECSIZE']
            if width % recsize != 0:
                raise VicarError(f'binary header length {width} is incompatible with '
                                 f'RECSIZE={recsize}')
            nlb = width // recsize

        # Tests passed
        self._binheader = None if width == 0 else value
        self._label['BHOST'] = _HOST
        self._label['NLB'] = nlb

        if isinstance(value, np.ndarray):
            if value.dtype.kind in 'cf':
                self._label['BREALFMT'] = VicarImage._realfmt(value)
            elif value.dtype.kind in 'ui' and value.dtype.itemsize > 1:
                self._label['BINTFMT'] = VicarImage._intfmt(value)

    ######################################################################################
    # Support for the Standard Python API
    ######################################################################################

    def copy(self):
        """A copy of this VicarImage.

        The copied label is independent of the original, but the data and prefix array
        are shared.

        Returns:
            VicarImage: The copy.
        """

        return VicarImage(source=self.label.copy(), array=self._array,
                          prefix=self._prefix, binheader=self._binheader)

    def deepcopy(self):
        """An independent (deep) copy of this VicarImage.

        Returns:
            VicarImage: The deep copy.
        """

        vim = self.copy()
        vim._array = None if self._array is None else self._array.copy()
        vim._prefix = None if self._prefix is None else self._prefix.copy()
        return vim

    def __eq__(self, other):
        """True if this VicarImage is equal to the given object.

        VicarImages are equal if the label content, data array, prefix array, binary
        header are all equal. Label formatting hints need not be the same.

        Parameters:
            other (VicarImage): Second VicarImage to compare with this.

        Returns:
            bool: True if the objects are equal.
        """

        if not isinstance(other, VicarImage):
            return False

        prefix1 = b'' if self._prefix is None else bytes(self._prefix)
        prefix2 = b'' if other._prefix is None else bytes(other._prefix)

        binheader1 = b'' if self._binheader is None else bytes(self._binheader)
        binheader2 = b'' if other._binheader is None else bytes(other._binheader)

        return (self._label == other._label
                and np.all(self._array == other._array)
                and prefix1 == prefix2
                and binheader1 == binheader2)

    def __str__(self):
        return str(self._label)

    def __repr__(self):
        return 'VicarImage("""' + self._label.as_string(sep='\n\n') + '""")'

    ######################################################################################
    # Public API
    ######################################################################################

    @staticmethod
    def from_file(filepath, extraneous='ignore', strict=True):
        """VicarImage object from an existing VICAR image file.

        Parameters:
            filepath (pathlib.Path or str): Path to an existing VICAR data file.

            extraneous (str, optional):
                How to handle the presence of extraneous bytes at the end of the file, one
                of:

                * "error"   to raise VicarError;
                * "warn"    to raise a UserWarning;
                * "print"   to print a message;
                * "ignore"  to ignore;
                * "include" to include the extraneous bytes as part of the return.

            strict (bool, optional):
                True (the default) to require strict conformance to the VICAR standard;
                False for a looser version of the standard.

        Returns:
            VicarImage or (VicarImage, bytes or None): A VicarImage or a tuple containing:

            * VicarImage: A new object containing the content of the specified file.
            * bytes or None: Any extraneous bytes from the end of the file, included if
              `extraneous` equals "include".
        """

        info = VicarImage._read_file(filepath, extraneous=extraneous, strict=strict)
        (label, data, prefix, binheader) = info[:4]
        vim = VicarImage(source=label, array=data, prefix=prefix, binheader=binheader)

        if extraneous == 'include':
            return (vim, info[4])

        return vim

    @staticmethod
    def from_array(array, strict=True):
        """Construct a VicarImage object for an array.

        Parameters:
            array (array-like): The data array to use in this VicarImage object.
            strict (bool, optional):
                True (the default) to require strict conformance to the VICAR standard;
                False for a looser version of the standard.

        Returns:
            VicarImage: A new VicarImage object containing this data array.
        """

        return VicarImage(source=[], array=array, strict=strict)

    def write_file(self, filepath=None):
        """Write the VicarImage object into a file.

        Parameters:
            filepath (path.Pathlib or str, optional):
                Optional the path of the file to write. If not specified but this object's
                filepath attribute is defined, it will write to this file.
        """

        filepath = filepath or self.filepath

        if self._array is None:
            raise VicarError('Image array is missing for ' + str(filepath))

        self.filepath = filepath

        # Open the file for binary write
        with self._filepath.open('wb') as f:

            labels = self._label.export(resize=True)
            f.write(labels[0].encode('latin8'))

            if self._binheader is not None:
                if isinstance(self._binheader, np.ndarray):
                    f.write(self._binheader.data)
                else:
                    f.write(self._binheader)

            if self._prefix is None and self._array is not None:
                f.write(self._array.data)
            else:
                n2 = self._label['N2']
                n3 = self._label['N3']
                nbb = self._label['NBB']
                recsize = self._label['RECSIZE']
                array = np.empty((n3,n2,recsize), dtype='uint8')
                array[:,:,:nbb] = self._prefix.view(dtype='uint8')
                array[:,:,nbb:] = self._array.view(dtype='uint8')
                f.write(array.data)

            f.write(labels[1].encode('latin8'))

    def binheader_array(self, kind='', size=None):
        """The numbers embedded in a binary header.

        This method is capable of reading "ISIS" table files when those tables consist
        entirely of a single data format. It uses the FMT_DEFAULT parameter to determine
        dtype, and uses the NR and NC values, if present, to determine the number of rows
        and columns in the table.

        Parameters:
            kind (str, optional):
                Optional single-letter code for the data type: "u" for unsigned int; "i"
                for signed int; "f" for float. If not specified, the kind is inferred from
                the value of the FMT_DEFAULT parameter.
            size (int, optional):
                Number of bytes per value. If not provided, it is inferred from the
                FMT_DEFAULT parameter if present. Otherwise, the default is 1 for kind =
                "u"; 2 for kind = "i"; 4 for kind = "f".

        Returns:
            np.ndarray or None:
                The binary header as an array; None if there is no binary header.
        """

        if self._binheader is None:
            return None

        if isinstance(self._binheader, np.ndarray):
            return self._binheader

        label = self._label

        # Determine kind, size, and dtype
        if not kind:
            dpref = ('>' if label['BREALFMT'] == 'IEEE' else '<')
            dtype = dpref + _DTYPE_FROM_FORMAT[label.get('FMT_DEFAULT', 'REAL')]
            kind = dtype[1]
            size = int(dtype[2:])

        elif kind in 'ui':
            size = size if size else 2 if kind == 'i' else 1
            dpref = ('<' if label['BINTFMT'] == 'LOW' else '>')
            dtype = dpref + kind + str(size)

        else:
            size = size if size else 4
            dpref = ('>' if label['BREALFMT'] == 'IEEE' else '<')
            dtype = dpref + kind + str(size)

        # Convert the bytes to an array
        values = np.frombuffer(self._binheader, dtype=dtype)

        # Check for table dimensions and them if found
        if 'NR' in label and 'NC' in label:
            nr = label['NR']
            nc = label['NC']
            values = values[:(nr*nc)]
            values = values.reshape(nr,nc)

        # Deal with the possibility of Vax reals
        if kind not in 'ui' and label['BREALFMT'] == 'VAX':
            if size == 8:       # pragma: no cover
                values = vax.from_vax64(values)
            else:
                values = vax.from_vax32(values)

        return values

    ######################################################################################
    # File I/O
    ######################################################################################

    @staticmethod
    def _read_file(filepath, *, extraneous='ignore', strict=True):
        """The VICAR data array, binary header, and prefix bytes from the specified
        data file.

        Parameters:
            filepath (pathlib.Path or str): Path to an existing VICAR data file.

            extraneous (str, optional):
                How to handle the presence of extraneous bytes at the end of the file, one
                of:

                * "error"   to raise VicarError;
                * "warn"    to raise a UserWarning;
                * "print"   to print a message;
                * "ignore"  to ignore;
                * "include" to include the extraneous bytes as part of the return.

            strict (bool, optional): True (the default) to require strict conformance to
                the VICAR standard; False for a looser version of the standard.

        Returns:
            (VicarLabel, np.ndarray, np.ndarray or None, bytes[, bytes]):
            A tuple containing:

            * VicarLabel: The label.
            * np.ndarray: The data as a 3D array converted to native format.
            * np.ndarray or None: The prefix array as a 3D array of unsigned bytes if
              present; otherwise, None.
            * bytes: The binary header if present; otherwise, None.
            * bytes, optional: Any extraneous bytes at the end of the file, included if
              `extraneous` equals "include".

        Raises:
            OSError: If the referenced file could not be read.
            ValueError: If `extraneous` does not have a valid value.
            VicarError: If the referenced file does not conform to the VICAR standard.
        """

        if extraneous not in ('ignore', 'print', 'warn', 'error', 'include'):
            raise ValueError('invalid input value for extraneous: ' + repr(extraneous))

        filepath = pathlib.Path(filepath)
        with filepath.open('rb') as f:

            # Get the label
            (label, extra) = VicarLabel.read_label(f, _extra=True)

            # Handle extraneous bytes
            if extra:
                if all(c == 0 for c in extra):
                    message = (f'{filepath} has {len(extra)} zero-valued trailing bytes')
                else:       # pragma: no cover
                    message = (f'{filepath} has {len(extra)} trailing bytes')

                if extraneous == 'print':
                    print(message)
                elif extraneous == 'warn':
                    warnings.warn(message)
                elif extraneous == 'error':
                    raise VicarError(message)

            # Extract key label parameters
            ldict = VicarLabel(label, strict=strict)
            lblsize = ldict['LBLSIZE']              # bytes in header
            recsize = ldict['RECSIZE']              # bytes per record
            nlb     = ldict['NLB']                  # records of binary header
            nbb     = ldict['NBB']                  # number of binary prefix bytes
            intfmt  = ldict.get('INTFMT', 'LOW')    # LOW or HIGH
            realfmt = ldict.get('REALFMT', 'VAX')   # IEEE, RIEEE, or VAX
            format_ = ldict['FORMAT']               # BYTE, HALF, FULL, REAL, etc.

            # Read the binary header
            f.seek(lblsize)
            if nlb:
                binheader = f.read(nlb * recsize)
            else:
                binheader = None

            # Read the data and prefix bytes
            ldict._n123_from_nbls()     # Sometimes N1, N2, N3 are wrong
            n2 = ldict['N2']
            n3 = ldict['N3']
            if n2 and n3:
                data = np.frombuffer(f.read(n3 * n2 * recsize), dtype='uint8')
                data = data.reshape(n3, n2, recsize)
            else:
                data = None

        # Separate the prefix bytes from the data
        if nbb and data is not None:
            array = data[:,:,nbb:].copy()
            prefix = data[:,:,:nbb].copy()
        else:
            array = data
            prefix = None

        # Convert the array to native format
        if array is not None:
            dtype = _DTYPE_FROM_FORMAT[format_]
            if dtype[0] in 'ui':
                dtype = ('>' if intfmt == 'HIGH' else '<') + dtype
            else:       # "fc"
                if realfmt == 'VAX':        # pragma: no cover
                    if dtype[-1] == '8':
                        array = vax.from_vax64(array)
                    else:
                        array = vax.from_vax32(array)
                    dtype = '<' + dtype
                else:
                    dtype = ('>' if realfmt == 'IEEE' else '<') + dtype

            array = array.view(dtype=dtype)                     # define actual format
            array = np.asarray(array, dtype='=' + dtype[1:])    # convert to native format

        if extraneous == 'include':
            return (ldict, array, prefix, binheader, extra)

        return (ldict, array, prefix, binheader)

    ######################################################################################
    # Public methods inherited from the VicarLabel object
    ######################################################################################

    def __len__(self):
        """The number of keywords in the label of this VicarImage.

        Returns:
            int: Number of parameters in the VICAR label
        """

        return len(self._label)

    def __getitem__(self, key):
        """Retrieve the value or values of the VICAR parameter defined by key, using
        various indexing options.

        Parameters:
            key (int, name, or tuple): The indexing key, interpreted as follows:

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

                Append a "+" to the name to return a list of all values where the
                constraints are satisfied, starting with the first or "nth".

        Returns:
            int, float, string, or list:
                The value of the indexed parameter if there is no name ending in "+";
                otherwise, a list of all the parameter values starting with the "nth".

        Raises:
            IndexError: If any numeric component of the key is out of range.
            KeyError: If the parameter name is not present in the label (or the section of
                the label defined by `after_name` and `after_value`), or if the key format
                is unrecognized.
            TypeError: If the key is not a recognized type or contains a component that is
                not of a recognized type.
            ValueError: If no identified parameter equals `value` or if no occurrence of
                `after_name` equals `after_value`.
        """

        return self._label.__getitem__(key)

    def get(self, key, default):
        """Retrieve the value of the VICAR parameter defined by the given key, using
        various indexing options.

        If the key is not found, return a specified default value.

        Parameters:
            key (int, str, or tuple): The indexing key, interpreted as follows:

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

                Append a "+" to the name to return a list of all values where the
                constraints are satisfied, starting with the first or "nth".

            default (int, float, str, or list): The value to return if the key is not
                found.

        Returns:
            int, float, str, or list:
                If a name is provided that ends in a plus, the returned value will be the
                list of all values of the selected key, or else `[default]` if the list
                would be empty or the key would raise an error.

                Otherwise, the returned value is that of the key if present, or `default`
                if it is not.

        Raises:
            IndexError: If any numeric component of the key is out of range.
            KeyError: If the parameter name is not present in the label (or the section of
                the label defined by `after_name` and `after_value`), or if the key format
                is unrecognized.
            TypeError: If the key is not a recognized type or contains a component that is
                not of a recognized type.
            ValueError: If no identified parameter equals `value` or if no occurrence of
                `after_name` equals `after_value`.
        """

        return self._label.get(key, default)

    def __setitem__(self, key, value):
        """Set the value of the VICAR parameter defined by the given key; define a new
        parameter name and value if necessary.

        Parameters:
            key (int, str, or tuple): The indexing key, interpreted as follows:

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

                Append a "+" to the name to force a new occurrence of the key to be
                inserted, even if the key already exists.

            value (int, float, string, list, or tuple): Value to assign to the indexed
                entry in the label.

                Optional formatting can be included if a user wants additional control
                over how this value will be formatted in the label string will be
                formatted, by replacing the value with a tuple (value, hints...). See
                ``Notes`` for details about formatting hints.

        Raises:
            IndexError: If any numeric component of the key is out of range.
            KeyError: If the parameter name is not present in the label (or the section of
                the label defined by `after_name` and `after_value`), or if the key format
                is unrecognized.
            TypeError: If the key is not a recognized type or contains a component that is
                not of a recognized type.
            ValueError: If no identified parameter equals `value` or if no occurrence of
                `after_name` equals `after_value`.

        Notes:
            Formatting hints can be included wherever a VICAR parameter value is
            specified; simply replace the value by a tuple (value, hints...).

            Hints can be specified using up to four items:

                ([`format`][[[, `name_blanks`], `val_blanks`], `sep_blanks`])

            where:

                * `format` is a format string, e.g., "%+7d" or "%7.3f".
                * `name_blanks` is the number of blank characters after the name and
                  before the equal sign; zero by default.
                * `val_blanks` is the number of blank characters after the equal sign and
                  before the value; zero by default.
                * `sep_blanks` is the number of blanks after the value and before the next
                  label parameter or the label's end; two by default.

            Note the use of square brackets in the tuple expression above. If the first
            hint value is a string, it is interpreted as `format`; otherwise, the `format`
            is unspecified. After the optional format, values are interpreted as numbers
            of blanks. If only one int is provided, it defines `sep_blanks`, with
            `val_blanks` and `name_blanks` set to zero. If two trailing ints are provided,
            they define `val_blanks` and `sep_blanks`, with `name_blanks` set to zero.

            For example, if the name is "TEXP" and the value is 1.5, this is how hint
            values are interpreted::

                <no hints>        = ("", 0, 0, 0)     -> "TEXP=1.5  "
                "%.3f"            = ("%.3f", 0, 0, 0) -> "TEXP=1.500  "
                ("%.3f", 4)       = ("%.3f", 0, 0, 4) -> "TEXP=1.500    "
                ("%.3f", 1, 4)    = ("%.3f", 0, 1, 4) -> "TEXP= 1.500    "
                ("%.3f", 2, 1, 4) = ("%.3f", 2, 1, 4) -> "TEXP  = 1.500    "
                4                 = ("", 0, 0, 4)     -> "TEXP=1.5    "
                (1, 4)            = ("", 0, 1, 4)     -> "TEXP= 1.5    "
                (2, 1, 4)         = ("", 2, 1, 4)     -> "TEXP  = 1.5    "

            When the parameter value is a list, it is also possible to embed formatting
            information on an item by item basis. Replace any item value by a tuple:

                (`item` [, `format`][[, `blanks_before`], `blanks_after`])

            where:

                * `format` is a format string, e.g., "%+07d", "%12.3e" or "%.4f".
                * `blanks_before` is the number of blanks before the value, after the left
                  parenthesis or comma; zero is the default.
                * `blanks_after` is the number of blanks after the value and before the
                  next comma or the right parenthesis; zero is the default.

            Here are some examples of a list with embedded formatting for a parameter
            named "XY" with a value [7,6]::

                [(7, "%+02d"), 6]       = [(7, "%+02d", 0, 0), 6] -> "XY=[+07, 6]  "
                [(7, 2), 6]             = [(7, "", 0, 2), 6]      -> "XY=[7  , 6]  "
                [(7, 1, 2), 6])         = [(7, "", 1, 2), 6]      -> "XY=[ 7  , 6]  "
                [(7, "%02d", 2), 6])    = [(7, "%02d", 0, 2), 6]  -> "XY=[07  , 6]  "
                [(7, "%02d", 1, 2), 6]) = [(7, "%02d", 1, 2), 6]  -> "XY=[ 07  , 6]  "
        """

        name = self._label._get_name(key)
        has_plus = VicarLabel._has_plus(key)
        if not has_plus and name in _IMMUTABLE:
            indx = self._label.arg(key)
            if indx == self._label._key_index[name][0]:      # if first occurrence
                raise VicarError(f'VICAR parameter {name} cannot be modified')

        self._label.__setitem__(key, value)

    def __delitem__(self, key):
        """Delete the value of the VICAR parameter identified by the given key.

        Parameters:
            key (int, str, or tuple): The indexing key, interpreted as follows:

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

                Append a "+" to `name` to delete all of the label parameters whose names
                match the constraints, starting with the first or "nth".

        Raises:
            IndexError: If any numeric component of the key is out of range.
            KeyError: If the parameter name is not present in the label (or the section
                of the label defined by `after_name` and `after_value`), or if the key
                format is unrecognized.
            TypeError: If the key is not a recognized type or contains a component that is
                not of a recognized type.
            ValueError: If no identified parameter equals `value` or if no occurrence of
                `after_name` equals `after_value`.
        """

        name = self._label._get_name(key)
        if name in _IMMUTABLE:
            indx = self._label.arg(VicarLabel._remove_plus(key))
            if indx == self.label._key_index[name][0]:      # if first occurrence
                raise VicarError(f'VICAR parameter {name} cannot be deleted')

        self._label.__delitem__(key)

    def __contains__(self, key):
        """True if the given key can is found in the label of this VicarImage.

        Parameters:
            key (int, str, or tuple): The key identifying the label parameter to check.

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

        Returns:
            bool: True if the key is found within the label.
        """

        return self._label.__contains__(key)

    def arg(self, key):
        """The index or indices of the keyed item in the label of this VicarImage.

        Parameters:
            key (int, name, or tuple): The indexing key, interpreted as follows:

                * *int* = `n`: The "nth" parameter in the label. `n` can be positive or
                  negative.
                * *str* = `name`: The first occurrence in the label of a parameter with
                  this name.
                * (*str*, *int*) = (`name`, `n`): The "nth" occurrence in the label of a
                  parameter with this name. `n` can be positive or negative.
                * (*str*, *str*) = (`name`, `after`): The first occurrence of parameter
                  `name` after the first occurrence of parameter `after` and before the
                  second occurrence of `after`.
                * (*str*, *str*, *any*) = (`name`, `after`, `value`): The first occurrence
                  of parameter `name` after the first location where `after` equals
                  `value` and before the next occurrence of `after`.

                The last two options make it easy to reference a VICAR label parameter
                that is repeated. For example, if `image` is a VicarImage, then
                "`image['DAT_TIM', 'TASK', 'COPY']`" uniquely identifies the occurrence of
                `DAT_TIM` applicable to `TASK='COPY'` when there might be other `TASK`
                sections of the label containing other values of `DAT_TIM`.

                Append a "+" to the name to return a list of all indices where the
                constraints are satisfied, starting with the first or "nth".

            value (int, float, or string, optional):
                If provided, the identified parameter must equal this value. For an
                integer key, if the indexed parameter does not have this value, ValueError
                is raised. For any key involving a name, values of the named parameter
                that do not match `value` are skipped over until the one(s) with the
                correct value are found.

        Returns:
            int or list[int]:
                If "+" is appended to `name`, the list of all positive indices that
                identify matching parameters, starting with the "nth". Otherwise, the
                single positive index into the label identifying the parameter.

        Raises:
            IndexError: If any numeric component of the key is out of range.
            KeyError: If the parameter name is not present in the label (or the section of
                the label defined by `after_name` and `after_value`), or if the key format
                is unrecognized.
            TypeError: If the key is not a recognized type or contains a component that is
                not of a recognized type.
            ValueError: If no identified parameter equals `value` or if no occurrence of
                `after_name` equals `after_value`.
        """

        return self._label.arg(key)

    def __iter__(self):
        """Iterator over the unique names or (name, occurrence) pairs in the label of
        this VicarImage.

        Returns:
            iterator:
                The parameter keys within this label, in order. The key is the parameter
                name if it is unique or (name, occurrence number) otherwise.
        """

        return self._label.__iter__()

    def names(self, pattern=None):
        """Iterator for the VICAR parameter name strings in the label of this VicarImage.

        Parameters:
            pattern (str or re.Pattern, optional):
                Regular expression that can be used to filter the label parameter names.

        Returns:
            list: The matching parameter names within this label, in order.
        """

        return self._label.names(pattern=pattern)

    def keys(self, pattern=None):
        """Iterator over the keys in the label of this VicarImage. The key is the
        parameter name if it is unique or (name, occurrence number) otherwise.

        Parameters:
            pattern (str or re.Pattern, optional):
                Regular expression that can be used to filter the label parameter names.

        Returns:
            list:
                The parameter keys within this label, in order. Each key is the parameter
                name if it is unique or (name, occurrence number) otherwise.
        """

        return self._label.keys(pattern=pattern)

    def values(self, pattern=None):
        """Iterator over the values in the label of this VicarImage.

        Parameters:
            pattern (str or re.Pattern, optional):
                Regular expression that can be used to filter the label parameter names.

        Returns:
            iterator: The values of the matching parameters within this label, in order.
        """

        return self._label.values(pattern=pattern)

    def items(self, pattern=None, unique=True):
        """Iterator over the (key, value) pairs in the label of this VicarImage.

        Parameters:
            pattern (str or re.Pattern, optional):
                Regular expression that can be used to filter the label parameter names.
            unique (bool, optional):
                True to return unique keys, in which non-unique names are replaced by
                tuples (name, occurrence). If False, all keys are name strings, and a name
                may appear multiple times.

        Returns:
            iterator:
                Tuples (name, value) of the matching parameter names within this label, in
                order.
        """

        return self._label.items(pattern=pattern, unique=unique)

    def args(self, pattern=None):
        """Iterator over the numerical indices of the keywords in the label of this
        VicarImage.

        Parameters:
            pattern (str or re.Pattern, optional):
                Regular expression that can be used to filter the label parameter names.

        Returns:
            iterator:
                The indices of the matching parameter names within this label, in order.
        """

        return self._label.args(pattern=pattern)

    def as_dict(self):
        """The VicarLabel object.

        DEPRECATED, provided primarily for backward compatibility. Use the `label`
        property instead, because it behaves as a dictionary.

        Returns:
            VicarLabel: The VicarLabel object associated with this VicarImage.
        """

        return self._label

    ######################################################################################
    # Utilities
    ######################################################################################

    def _intfmt(x):
        """Determine the INTFMT based on the dtype of an array."""

        if isinstance(x, np.ndarray):
            if x.dtype.byteorder == '<':    # pragma: no cover
                return 'LOW'
            if x.dtype.byteorder == '>':    # pragma: no cover
                return 'HIGH'

        return 'LOW' if sys.byteorder == 'little' else 'HIGH'

    def _realfmt(x):
        """Determine the REALFMT value based on the dtype of an array."""

        if isinstance(x, np.ndarray):
            if x.dtype.byteorder == '>':    # pragma: no cover
                return 'IEEE'
            if x.dtype.byteorder == '<':    # pragma: no cover
                return 'RIEEE'

        return 'RIEEE' if sys.byteorder == 'little' else 'IEEE'

    def _format_isint(x):
        """True if this array contains integer values."""

        try:
            key = x.dtype.kind + str(x.itemsize)
            return _FORMAT_FROM_DTYPE[key]
        except KeyError:
            raise VicarError(f'array dtype "{x.dtype}" is not supported by VICAR')

    def _check_array_vs_prefix(array, prefix):
        """Raise an exception if the given image array and prefix byte array are not
        compatible.

        Parameters:
            array (np.ndarray or None): A data array.
            prefix (np.ndarray or None): A prefix array.

        Raises:
            VicarError: If the arrays are invalid or inconsistent with one another.
        """

        if array is not None and array.ndim != 3:
            raise VicarError(f'data array shape {array.shape} is not 2-D or 3-D')

        if prefix is not None and prefix.ndim != 3:
            raise VicarError(f'prefix array shape {prefix.shape} is not 2-D or 3-D')

        if array is None or prefix is None:
            return

        if array.shape[:-1] != prefix.shape[:-1]:
            raise VicarError('data and prefix arrays have incompatible shapes: '
                             f'{array.shape}, {prefix.shape}')

        (format1, isint1) = VicarImage._format_isint(array)
        if isint1:
            intfmt1 = VicarImage._intfmt(array)
        else:
            realfmt1 = VicarImage._realfmt(array)

        (format2, isint2) = VicarImage._format_isint(prefix)
        if isint2:
            intfmt2 = VicarImage._intfmt(prefix)
        else:
            realfmt2 = VicarImage._realfmt(prefix)

        if isint1 and isint2:
            if format1 != 'BYTE' and format2 != 'BYTE' and intfmt1 != intfmt2:
                raise VicarError('data and prefix array formats are incompatible: '
                                 f'{intfmt1}, {intfmt2}')
        if not isint1 and not isint2:
            if realfmt1 != realfmt2:
                raise VicarError('data and prefix array formats are incompatible: '
                                 f'{realfmt1}, {realfmt2}')

##########################################################################################
